@@ -18,8 +18,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/resource.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <unistd.h>
 #include <wchar.h>
 #include <wctype.h>
@@ -34,8 +36,36 @@
 #  include <util.h>
 #endif
 
-bool is_nonexistence_error(int error) {
-	return error == ENOENT || errno == ENOTDIR;
+bool error_is_like(int error, int category) {
+	if (error == category) {
+		return true;
+	}
+
+	switch (category) {
+	case ENOENT:
+		return error == ENOTDIR;
+
+#if __DragonFly__
+	// https://twitter.com/tavianator/status/1742991411203485713
+	case ENAMETOOLONG:
+		return error == EFAULT;
+#endif
+	}
+
+	return false;
+}
+
+bool errno_is_like(int category) {
+	return error_is_like(errno, category);
+}
+
+int try(int ret) {
+	if (ret >= 0) {
+		return ret;
+	} else {
+		bfs_assert(errno > 0, "errno should be positive, was %d\n", errno);
+		return -errno;
+	}
 }
 
 char *xdirname(const char *path) {
@@ -282,8 +312,12 @@ const char *xstrerror(int errnum) {
 	const char *ret = NULL;
 	static thread_local char buf[256];
 
-#if __APPLE__
-	// No strerror_l() on macOS
+	// - __APPLE__
+	// - __COSMOPOLITAN__
+	//     - No strerror_l()
+	// - __FreeBSD__ && SANITIZE_MEMORY
+	//     - duplocale() triggers https://github.com/llvm/llvm-project/issues/65532
+#if __APPLE__ || __COSMOPOLITAN__ || (__FreeBSD__ && SANITIZE_MEMORY)
 	if (strerror_r(errnum, buf, sizeof(buf)) == 0) {
 		ret = buf;
 	}
@@ -302,10 +336,10 @@ const char *xstrerror(int errnum) {
 
 	if (copy != (locale_t)0) {
 		ret = strerror_l(errnum, copy);
-	}
 
-	if (loc == LC_GLOBAL_LOCALE) {
-		freelocale(copy);
+		if (loc == LC_GLOBAL_LOCALE) {
+			freelocale(copy);
+		}
 	}
 #endif
 
@@ -367,6 +401,38 @@ void xstrmode(mode_t mode, char str[11]) {
 	}
 }
 
+/** Check if an rlimit value is infinite. */
+static bool rlim_isinf(rlim_t r) {
+	// Consider RLIM_{INFINITY,SAVED_{CUR,MAX}} all equally infinite
+	if (r == RLIM_INFINITY) {
+		return true;
+	}
+
+#ifdef RLIM_SAVED_CUR
+	if (r == RLIM_SAVED_CUR) {
+		return true;
+	}
+#endif
+
+#ifdef RLIM_SAVED_MAX
+	if (r == RLIM_SAVED_MAX) {
+		return true;
+	}
+#endif
+
+	return false;
+}
+
+int rlim_cmp(rlim_t a, rlim_t b) {
+	bool a_inf = rlim_isinf(a);
+	bool b_inf = rlim_isinf(b);
+	if (a_inf || b_inf) {
+		return a_inf - b_inf;
+	}
+
+	return (a > b) - (a < b);
+}
+
 dev_t xmakedev(int ma, int mi) {
 #ifdef makedev
 	return makedev(ma, mi);
@@ -391,6 +457,14 @@ int xminor(dev_t dev) {
 #endif
 }
 
+pid_t xwaitpid(pid_t pid, int *status, int flags) {
+	pid_t ret;
+	do {
+		ret = waitpid(pid, status, flags);
+	} while (ret < 0 && errno == EINTR);
+	return ret;
+}
+
 int dup_cloexec(int fd) {
 #ifdef F_DUPFD_CLOEXEC
 	return fcntl(fd, F_DUPFD_CLOEXEC, 0);
@@ -563,8 +637,14 @@ error:
 int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *clear) {
 #if BSD && !__GNU__
 	char *str_arg = (char *)*str;
-	unsigned long set_arg = 0;
-	unsigned long clear_arg = 0;
+
+#if __OpenBSD__
+	typedef uint32_t bfs_fflags_t;
+#else
+	typedef unsigned long bfs_fflags_t;
+#endif
+	bfs_fflags_t set_arg = 0;
+	bfs_fflags_t clear_arg = 0;
 
 #if __NetBSD__
 	int ret = string_to_flags(&str_arg, &set_arg, &clear_arg);
@@ -586,18 +666,18 @@ int xstrtofflags(const char **str, unsigned long long *set, unsigned long long *
 #endif
 }
 
-/** mbrtowc() wrapper. */
-static int xmbrtowc(wchar_t *wc, size_t *i, const char *str, size_t len, mbstate_t *mb) {
-	size_t mblen = mbrtowc(wc, str + *i, len - *i, mb);
+wint_t xmbrtowc(const char *str, size_t *i, size_t len, mbstate_t *mb) {
+	wchar_t wc;
+	size_t mblen = mbrtowc(&wc, str + *i, len - *i, mb);
 	switch (mblen) {
 	case -1: // Invalid byte sequence
 	case -2: // Incomplete byte sequence
 		*i += 1;
-		memset(mb, 0, sizeof(*mb));
-		return -1;
+		*mb = (mbstate_t){0};
+		return WEOF;
 	default:
 		*i += mblen;
-		return 0;
+		return wc;
 	}
 }
 
@@ -605,16 +685,14 @@ size_t xstrwidth(const char *str) {
 	size_t len = strlen(str);
 	size_t ret = 0;
 
-	mbstate_t mb;
-	memset(&mb, 0, sizeof(mb));
-
+	mbstate_t mb = {0};
 	for (size_t i = 0; i < len;) {
-		wchar_t wc;
-		if (xmbrtowc(&wc, &i, str, len, &mb) == 0) {
-			ret += wcwidth(wc);
-		} else {
+		wint_t wc = xmbrtowc(str, &i, len, &mb);
+		if (wc == WEOF) {
 			// Assume a single-width '?'
 			++ret;
+		} else {
+			ret += xwcwidth(wc);
 		}
 	}
 
@@ -634,27 +712,18 @@ static unsigned char ctype_cache[UCHAR_MAX + 1];
 
 /** Initialize the ctype cache. */
 static void char_cache_init(void) {
-#if __FreeBSD__ && SANITIZE_MEMORY
-// Work around https://github.com/llvm/llvm-project/issues/65532
-#  define bfs_isprint (isprint)
-#  define bfs_isspace (isspace)
-#else
-#  define bfs_isprint isprint
-#  define bfs_isspace isspace
-#endif
-
 	for (size_t c = 0; c <= UCHAR_MAX; ++c) {
-		if (bfs_isprint(c)) {
+		if (xisprint(c)) {
 			ctype_cache[c] |= IS_PRINT;
 		}
-		if (bfs_isspace(c)) {
+		if (xisspace(c)) {
 			ctype_cache[c] |= IS_SPACE;
 		}
 	}
 }
 
 /** Check if a character is printable. */
-static bool xisprint(unsigned char c, enum wesc_flags flags) {
+static bool wesc_isprint(unsigned char c, enum wesc_flags flags) {
 	if (ctype_cache[c] & IS_PRINT) {
 		return true;
 	}
@@ -670,21 +739,12 @@ static bool xisprint(unsigned char c, enum wesc_flags flags) {
 }
 
 /** Check if a wide character is printable. */
-static bool xiswprint(wchar_t c, enum wesc_flags flags) {
-#if __FreeBSD__ && SANITIZE_MEMORY
-// Work around https://github.com/llvm/llvm-project/issues/65532
-#  define bfs_iswprint (iswprint)
-#  define bfs_iswspace (iswspace)
-#else
-#  define bfs_iswprint iswprint
-#  define bfs_iswspace iswspace
-#endif
-
-	if (bfs_iswprint(c)) {
+static bool wesc_iswprint(wchar_t c, enum wesc_flags flags) {
+	if (xiswprint(c)) {
 		return true;
 	}
 
-	if (!(flags & WESC_SHELL) && bfs_iswspace(c)) {
+	if (!(flags & WESC_SHELL) && xiswspace(c)) {
 		return true;
 	}
 
@@ -708,32 +768,31 @@ static size_t printable_len(const char *str, size_t len, enum wesc_flags flags)
 		}
 
 		for (size_t j = 0; j < sizeof(word); ++i, ++j) {
-			if (!xisprint(str[i], flags)) {
+			if (!wesc_isprint(str[i], flags)) {
 				return i;
 			}
 		}
 	}
 
 	for (; i < len; ++i) {
-		unsigned char c = str[i];
-		if (!isascii(c)) {
+		char c = str[i];
+		if (!xisascii(c)) {
 			goto multibyte;
 		}
-		if (!xisprint(c, flags)) {
+		if (!wesc_isprint(c, flags)) {
 			return i;
 		}
 	}
 
-	mbstate_t mb;
-multibyte:
-	memset(&mb, 0, sizeof(mb));
+multibyte:;
+	mbstate_t mb = {0};
 
 	for (size_t j = i; i < len; i = j) {
-		wchar_t wc;
-		if (xmbrtowc(&wc, &j, str, len, &mb) != 0) {
+		wint_t wc = xmbrtowc(str, &j, len, &mb);
+		if (wc == WEOF) {
 			break;
 		}
-		if (!xiswprint(wc, flags)) {
+		if (!wesc_iswprint(wc, flags)) {
 			break;
 		}
 	}
@@ -774,16 +833,14 @@ static const char *dollar_esc(char c) {
 static char *dollar_quote(char *dest, char *end, const char *str, size_t len, enum wesc_flags flags) {
 	dest = xstpecpy(dest, end, "$'");
 
-	mbstate_t mb;
-	memset(&mb, 0, sizeof(mb));
-
+	mbstate_t mb = {0};
 	for (size_t i = 0; i < len;) {
 		size_t start = i;
 		bool safe = false;
 
-		wchar_t wc;
-		if (xmbrtowc(&wc, &i, str, len, &mb) == 0) {
-			safe = xiswprint(wc, flags);
+		wint_t wc = xmbrtowc(str, &i, len, &mb);
+		if (wc != WEOF) {
+			safe = wesc_iswprint(wc, flags);
 		}
 
 		for (size_t j = start; safe && j < i; ++j) {
@@ -816,7 +873,7 @@ static char *dollar_quote(char *dest, char *end, const char *str, size_t len, en
 /** How much of this string is safe as a bare word? */
 static size_t bare_len(const char *str, size_t len) {
 	// https://pubs.opengroup.org/onlinepubs/9699919799/utilities/V3_chap02.html#tag_18_02
-	size_t ret = strcspn(str, "|&;<>()$`\\\"' *?[#˜=%!");
+	size_t ret = strcspn(str, "|&;<>()$`\\\"' *?[#˜=%!{}");
 	return ret < len ? ret : len;
 }
 
