@@ -26,15 +26,11 @@
  *
  * Pushes are implemented with an unconditional
  *
- *     fetch_add(&ioqq->head, IOQ_STRIDE)
+ *     fetch_add(&ioqq->head, 1)
  *
  * which scales better on many architectures than compare-and-swap (see [1] for
- * details).  Pops are implemented similarly.  We add IOQ_STRIDE rather than 1
- * so that successive queue elements are on different cache lines, but the
- * exposition below uses 1 for simplicity.
- *
- * Since the fetch-and-adds are unconditional, non-blocking readers can get
- * ahead of writers:
+ * details).  Pops are implemented similarly.  Since the fetch-and-adds are
+ * unconditional, non-blocking readers can get ahead of writers:
  *
  *     Reader              Writer
  *     ────────────────    ──────────────────────
@@ -204,17 +200,6 @@ struct ioqq {
 	cache_align ioq_slot slots[];
 };
 
-// If we assign slots sequentially, threads will likely be operating on
-// consecutive slots.  If these slots are in the same cache line, that will
-// result in false sharing.  We can mitigate this by assigning slots with a
-// stride larger than a cache line e.g. 0, 9, 18, ..., 1, 10, 19, ...
-// As long as the stride is relatively prime to circular buffer length, we'll
-// still use every available slot.  Since the length is a power of two, that
-// means the stride must be odd.
-
-#define IOQ_STRIDE ((FALSE_SHARING_SIZE / sizeof(ioq_slot)) | 1)
-bfs_static_assert(IOQ_STRIDE % 2 == 1);
-
 /** Destroy an I/O command queue. */
 static void ioqq_destroy(struct ioqq *ioqq) {
 	if (!ioqq) {
@@ -354,17 +339,6 @@ static bool ioq_slot_push(struct ioqq *ioqq, ioq_slot *slot, struct ioq_ent *ent
 	return !(prev & IOQ_SKIP);
 }
 
-/** Push an entry onto the queue. */
-static void ioqq_push(struct ioqq *ioqq, struct ioq_ent *ent) {
-	while (true) {
-		size_t i = fetch_add(&ioqq->head, IOQ_STRIDE, relaxed);
-		ioq_slot *slot = &ioqq->slots[i & ioqq->slot_mask];
-		if (ioq_slot_push(ioqq, slot, ent)) {
-			break;
-		}
-	}
-}
-
 /** (Try to) pop an entry from a slot. */
 static struct ioq_ent *ioq_slot_pop(struct ioqq *ioqq, ioq_slot *slot, bool block) {
 	uintptr_t prev = load(slot, relaxed);
@@ -398,13 +372,80 @@ static struct ioq_ent *ioq_slot_pop(struct ioqq *ioqq, ioq_slot *slot, bool bloc
 	return (struct ioq_ent *)(prev << 1);
 }
 
+/** Push an entry onto the queue. */
+static void ioqq_push(struct ioqq *ioqq, struct ioq_ent *ent) {
+	while (true) {
+		size_t i = fetch_add(&ioqq->head, 1, relaxed);
+		ioq_slot *slot = &ioqq->slots[i & ioqq->slot_mask];
+		if (ioq_slot_push(ioqq, slot, ent)) {
+			break;
+		}
+	}
+}
+
+/** Push a batch of entries to the queue. */
+static void ioqq_push_batch(struct ioqq *ioqq, struct ioq_ent *batch[], size_t size) {
+	size_t mask = ioqq->slot_mask;
+	do {
+		size_t i = fetch_add(&ioqq->head, size, relaxed);
+		for (size_t j = i + size; i != j; ++i) {
+			ioq_slot *slot = &ioqq->slots[i & mask];
+			if (ioq_slot_push(ioqq, slot, *batch)) {
+				++batch;
+				--size;
+			}
+		}
+	} while (size > 0);
+}
+
 /** Pop an entry from the queue. */
 static struct ioq_ent *ioqq_pop(struct ioqq *ioqq, bool block) {
-	size_t i = fetch_add(&ioqq->tail, IOQ_STRIDE, relaxed);
+	size_t i = fetch_add(&ioqq->tail, 1, relaxed);
 	ioq_slot *slot = &ioqq->slots[i & ioqq->slot_mask];
 	return ioq_slot_pop(ioqq, slot, block);
 }
 
+/** Pop a batch of entries from the queue. */
+static void ioqq_pop_batch(struct ioqq *ioqq, struct ioq_ent *batch[], size_t size, bool block) {
+	size_t mask = ioqq->slot_mask;
+	size_t i = fetch_add(&ioqq->tail, size, relaxed);
+	for (size_t j = i + size; i != j; ++i) {
+		ioq_slot *slot = &ioqq->slots[i & mask];
+		*batch++ = ioq_slot_pop(ioqq, slot, block);
+		block = false;
+	}
+}
+
+/** Use cache-line-sized batches. */
+#define IOQ_BATCH (FALSE_SHARING_SIZE / sizeof(ioq_slot))
+
+/**
+ * A batch of entries to send all at once.
+ */
+struct ioq_batch {
+	/** The current batch size. */
+	size_t size;
+	/** The array of entries. */
+	struct ioq_ent *entries[IOQ_BATCH];
+};
+
+/** Send the batch to a queue. */
+static void ioq_batch_flush(struct ioqq *ioqq, struct ioq_batch *batch) {
+	if (batch->size > 0) {
+		ioqq_push_batch(ioqq, batch->entries, batch->size);
+		batch->size = 0;
+	}
+}
+
+/** An an entry to a batch, flushing if necessary. */
+static void ioq_batch_push(struct ioqq *ioqq, struct ioq_batch *batch, struct ioq_ent *ent) {
+	if (batch->size >= IOQ_BATCH) {
+		ioq_batch_flush(ioqq, batch);
+	}
+
+	batch->entries[batch->size++] = ent;
+}
+
 /** Sentinel stop command. */
 static struct ioq_ent IOQ_STOP;
 
@@ -461,7 +502,6 @@ static bool ioq_check_cancel(struct ioq *ioq, struct ioq_ent *ent) {
 	}
 
 	ent->result = -EINTR;
-	ioqq_push(ioq->ready, ent);
 	return true;
 }
 
@@ -496,12 +536,6 @@ static void ioq_dispatch_sync(struct ioq *ioq, struct ioq_ent *ent) {
 	ent->result = -ENOSYS;
 }
 
-/** Complete a single request synchronously. */
-static void ioq_complete(struct ioq *ioq, struct ioq_ent *ent) {
-	ioq_dispatch_sync(ioq, ent);
-	ioqq_push(ioq->ready, ent);
-}
-
 #if BFS_USE_LIBURING
 
 /** io_uring worker state. */
@@ -516,26 +550,10 @@ struct ioq_ring_state {
 	size_t submitted;
 	/** Whether to stop the loop. */
 	bool stop;
+	/** A batch of ready entries. */
+	struct ioq_batch ready;
 };
 
-/** Pop a request for ioq_ring_prep(). */
-static struct ioq_ent *ioq_ring_pop(struct ioq_ring_state *state) {
-	if (state->stop) {
-		return NULL;
-	}
-
-	// Block if we have nothing else to do
-	bool block = !state->prepped && !state->submitted;
-	struct ioq_ent *ret = ioqq_pop(state->ioq->pending, block);
-
-	if (ret == &IOQ_STOP) {
-		state->stop = true;
-		ret = NULL;
-	}
-
-	return ret;
-}
-
 /** Dispatch a single request asynchronously. */
 static struct io_uring_sqe *ioq_dispatch_async(struct io_uring *ring, struct ioq_ent *ent) {
 	struct io_uring_sqe *sqe = NULL;
@@ -577,10 +595,16 @@ static struct io_uring_sqe *ioq_dispatch_async(struct io_uring *ring, struct ioq
 	return NULL;
 }
 
+/** Check if ioq_ring_reap() has work to do. */
+static bool ioq_ring_empty(struct ioq_ring_state *state) {
+	return !state->prepped && !state->submitted && !state->ready.size;
+}
+
 /** Prep a single SQE. */
 static void ioq_prep_sqe(struct ioq_ring_state *state, struct ioq_ent *ent) {
 	struct ioq *ioq = state->ioq;
 	if (ioq_check_cancel(ioq, ent)) {
+		ioq_batch_push(ioq->ready, &state->ready, ent);
 		return;
 	}
 
@@ -589,24 +613,45 @@ static void ioq_prep_sqe(struct ioq_ring_state *state, struct ioq_ent *ent) {
 		io_uring_sqe_set_data(sqe, ent);
 		++state->prepped;
 	} else {
-		ioq_complete(ioq, ent);
+		ioq_dispatch_sync(ioq, ent);
+		ioq_batch_push(ioq->ready, &state->ready, ent);
 	}
 }
 
 /** Prep a batch of SQEs. */
 static bool ioq_ring_prep(struct ioq_ring_state *state) {
-	struct io_uring *ring = state->ring;
-
-	while (io_uring_sq_space_left(ring)) {
-		struct ioq_ent *ent = ioq_ring_pop(state);
-		if (!ent) {
-			break;
-		}
-
-		ioq_prep_sqe(state, ent);
+	if (state->stop) {
+		return false;
 	}
 
-	return state->prepped || state->submitted;
+	struct ioq *ioq = state->ioq;
+	struct io_uring *ring = state->ring;
+	struct ioq_ent *pending[IOQ_BATCH];
+
+	while (io_uring_sq_space_left(ring) >= IOQ_BATCH) {
+		bool block = ioq_ring_empty(state);
+		ioqq_pop_batch(ioq->pending, pending, IOQ_BATCH, block);
+
+		bool any = false;
+		for (size_t i = 0; i < IOQ_BATCH; ++i) {
+			struct ioq_ent *ent = pending[i];
+			if (ent == &IOQ_STOP) {
+				ioqq_push(ioq->pending, &IOQ_STOP);
+				state->stop = true;
+				goto done;
+			} else if (ent) {
+				ioq_prep_sqe(state, ent);
+				any = true;
+			}
+		}
+
+		if (!any) {
+			break;
+		}
+	}
+
+done:
+	return !ioq_ring_empty(state);
 }
 
 /** Reap a single CQE. */
@@ -628,7 +673,7 @@ static void ioq_reap_cqe(struct ioq_ring_state *state, struct io_uring_cqe *cqe)
 			int fd = ent->result;
 			if (ioq_check_cancel(ioq, ent)) {
 				xclose(fd);
-				return;
+				goto push;
 			}
 
 			struct ioq_opendir *args = &ent->opendir;
@@ -636,6 +681,8 @@ static void ioq_reap_cqe(struct ioq_ring_state *state, struct io_uring_cqe *cqe)
 			if (ent->result >= 0) {
 				// TODO: io_uring_prep_getdents()
 				bfs_polldir(args->dir);
+			} else {
+				xclose(fd);
 			}
 
 			break;
@@ -654,11 +701,12 @@ static void ioq_reap_cqe(struct ioq_ring_state *state, struct io_uring_cqe *cqe)
 	}
 
 push:
-	ioqq_push(ioq->ready, ent);
+	ioq_batch_push(ioq->ready, &state->ready, ent);
 }
 
 /** Reap a batch of CQEs. */
 static void ioq_ring_reap(struct ioq_ring_state *state) {
+	struct ioq *ioq = state->ioq;
 	struct io_uring *ring = state->ring;
 
 	while (state->prepped) {
@@ -677,6 +725,8 @@ static void ioq_ring_reap(struct ioq_ring_state *state) {
 
 		ioq_reap_cqe(state, cqe);
 	}
+
+	ioq_batch_flush(ioq->ready, &state->ready);
 }
 
 /** io_uring worker loop. */
@@ -697,15 +747,29 @@ static void ioq_ring_work(struct ioq_thread *thread) {
 static void ioq_sync_work(struct ioq_thread *thread) {
 	struct ioq *ioq = thread->parent;
 
-	while (true) {
-		struct ioq_ent *ent = ioqq_pop(ioq->pending, true);
-		if (ent == &IOQ_STOP) {
-			break;
+	bool stop = false;
+	while (!stop) {
+		struct ioq_ent *pending[IOQ_BATCH];
+		ioqq_pop_batch(ioq->pending, pending, IOQ_BATCH, true);
+
+		struct ioq_batch ready;
+		ready.size = 0;
+
+		for (size_t i = 0; i < IOQ_BATCH; ++i) {
+			struct ioq_ent *ent = pending[i];
+			if (ent == &IOQ_STOP) {
+				ioqq_push(ioq->pending, &IOQ_STOP);
+				stop = true;
+				break;
+			} else if (ent) {
+				if (!ioq_check_cancel(ioq, ent)) {
+					ioq_dispatch_sync(ioq, ent);
+				}
+				ioq_batch_push(ioq->ready, &ready, ent);
+			}
 		}
 
-		if (!ioq_check_cancel(ioq, ent)) {
-			ioq_complete(ioq, ent);
-		}
+		ioq_batch_flush(ioq->ready, &ready);
 	}
 }
 
@@ -724,6 +788,75 @@ static void *ioq_work(void *ptr) {
 	return NULL;
 }
 
+/** Initialize io_uring thread state. */
+static int ioq_ring_init(struct ioq *ioq, struct ioq_thread *thread) {
+#if BFS_USE_LIBURING
+	struct ioq_thread *prev = NULL;
+	if (thread > ioq->threads) {
+		prev = thread - 1;
+	}
+
+	if (prev && prev->ring_err) {
+		thread->ring_err = prev->ring_err;
+		return -1;
+	}
+
+	// Share io-wq workers between rings
+	struct io_uring_params params = {0};
+	if (prev) {
+		params.flags |= IORING_SETUP_ATTACH_WQ;
+		params.wq_fd = prev->ring.ring_fd;
+	}
+
+	// Use a page for each SQE ring
+	size_t entries = 4096 / sizeof(struct io_uring_sqe);
+	thread->ring_err = -io_uring_queue_init_params(entries, &thread->ring, &params);
+	if (thread->ring_err) {
+		return -1;
+	}
+
+	if (!prev) {
+		// Limit the number of io_uring workers
+		unsigned int values[] = {
+			ioq->nthreads, // [IO_WQ_BOUND]
+			0,             // [IO_WQ_UNBOUND]
+		};
+		io_uring_register_iowq_max_workers(&thread->ring, values);
+	}
+#endif
+
+	return 0;
+}
+
+/** Destroy an io_uring. */
+static void ioq_ring_exit(struct ioq_thread *thread) {
+#if BFS_USE_LIBURING
+	if (thread->ring_err == 0) {
+		io_uring_queue_exit(&thread->ring);
+	}
+#endif
+}
+
+/** Create an I/O queue thread. */
+static int ioq_thread_create(struct ioq *ioq, struct ioq_thread *thread) {
+	thread->parent = ioq;
+
+	ioq_ring_init(ioq, thread);
+
+	if (thread_create(&thread->id, NULL, ioq_work, thread) != 0) {
+		ioq_ring_exit(thread);
+		return -1;
+	}
+
+	return 0;
+}
+
+/** Join an I/O queue thread. */
+static void ioq_thread_join(struct ioq_thread *thread) {
+	thread_join(thread->id, NULL);
+	ioq_ring_exit(thread);
+}
+
 struct ioq *ioq_create(size_t depth, size_t nthreads) {
 	struct ioq *ioq = ZALLOC_FLEX(struct ioq, threads, nthreads);
 	if (!ioq) {
@@ -748,34 +881,12 @@ struct ioq *ioq_create(size_t depth, size_t nthreads) {
 		goto fail;
 	}
 
+	ioq->nthreads = nthreads;
 	for (size_t i = 0; i < nthreads; ++i) {
-		struct ioq_thread *thread = &ioq->threads[i];
-		thread->parent = ioq;
-
-#if BFS_USE_LIBURING
-		struct ioq_thread *prev = i ? &ioq->threads[i - 1] : NULL;
-		if (prev && prev->ring_err) {
-			thread->ring_err = prev->ring_err;
-		} else {
-			// Share io-wq workers between rings
-			struct io_uring_params params = {0};
-			if (prev) {
-				params.flags |= IORING_SETUP_ATTACH_WQ;
-				params.wq_fd = prev->ring.ring_fd;
-			}
-
-			size_t entries = depth / nthreads;
-			if (entries < 16) {
-				entries = 16;
-			}
-			thread->ring_err = -io_uring_queue_init_params(entries, &thread->ring, &params);
-		}
-#endif
-
-		if (thread_create(&thread->id, NULL, ioq_work, thread) != 0) {
+		if (ioq_thread_create(ioq, &ioq->threads[i]) != 0) {
+			ioq->nthreads = i;
 			goto fail;
 		}
-		++ioq->nthreads;
 	}
 
 	return ioq;
@@ -901,9 +1012,7 @@ void ioq_free(struct ioq *ioq, struct ioq_ent *ent) {
 
 void ioq_cancel(struct ioq *ioq) {
 	if (!exchange(&ioq->cancel, true, relaxed)) {
-		for (size_t i = 0; i < ioq->nthreads; ++i) {
-			ioqq_push(ioq->pending, &IOQ_STOP);
-		}
+		ioqq_push(ioq->pending, &IOQ_STOP);
 	}
 }
 
@@ -915,11 +1024,7 @@ void ioq_destroy(struct ioq *ioq) {
 	ioq_cancel(ioq);
 
 	for (size_t i = 0; i < ioq->nthreads; ++i) {
-		struct ioq_thread *thread = &ioq->threads[i];
-		thread_join(thread->id, NULL);
-#if BFS_USE_LIBURING
-		io_uring_queue_exit(&thread->ring);
-#endif
+		ioq_thread_join(&ioq->threads[i]);
 	}
 
 	ioqq_destroy(ioq->ready);
