@@ -1,2 +1,2 @@
 # Regression test: don't segfault on missing command
-fail invoke_bfs basic -ok \;
+! invoke_bfs basic -ok \;
